/* Copyright 2011-2012, Georgia Institute of Technology. All rights reserved. */
package org.gtri.icl.iead.policy_guide.sunxacml;

import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.URI;
import java.util.Map.Entry;
import java.util.Properties;
import java.util.logging.Logger;

import javax.servlet.ServletContext;

import com.sun.xacml.AbstractPolicy;
import com.sun.xacml.PolicyMetaData;
import com.sun.xacml.VersionConstraints;
import com.sun.xacml.finder.PolicyFinder;
import com.sun.xacml.finder.PolicyFinderModule;
import com.sun.xacml.finder.PolicyFinderResult;
import com.sun.xacml.support.finder.PolicyReader;

/**
 * This class is used to resolve XACML policy references in the Sample Application.
 * An XML Properties configuration file must be made available to this module.
 * The configuration file must contain mappings from Policy ID to policy file name.
 * @author Erin.Reddick AT gtri.gatech.edu
 *
 */
public class ReferencePolicyFinderModule extends PolicyFinderModule {

	/**
	 * Used to write to the system log
	 */
	private static final Logger log = Logger.getLogger(ReferencePolicyFinderModule.class.getName());
	
	/**
	 * The path of the directory that contains the referenced policy files 
	 */
	private final String policyDirectoryPath;
	
	/**
	 * The servletContext is used to read files off the disk from a web application
	 */
	private final ServletContext servletContext;
	
	/**
	 * The Properties representation of the configuration file
	 */
	private final Properties configProps;

	/**
	 * The SunXACML PolicyFinder instance provided by the SunXACML library.
	 * This class needs to pass policyFinder back to the SunXACML library
	 * at certain points.
	 */
	private PolicyFinder policyFinder;

	/**
	 * Uses servletContext to load and parse the configuration file identified by configFilname.
	 * The configuration file and referenced policies must be in the directory identified by
	 * policyDirectoryPath
	 * @param servletContext the ServletContext of the encompassing Servlet
	 * @param policyDirectoryPath the path of the directory that contains
	 * the referenced policies and the configuration file
	 * @param configFilename the name of the configuration file
	 * @throws IOException
	 */
	public ReferencePolicyFinderModule(ServletContext servletContext, String policyDirectoryPath, String configFilename) throws IOException {
		this.servletContext = servletContext;
		this.policyDirectoryPath = policyDirectoryPath;
		
		String configPath = policyDirectoryPath + "/" + configFilename;
		log.info("configPath - " + configPath);
		InputStream configIs = servletContext.getResourceAsStream(configPath);
		configProps = new Properties();
		configProps.loadFromXML(configIs);
		for(Entry<Object, Object> entry : configProps.entrySet()) {
			log.info("key - " + entry.getKey().toString() + " - - value - " + entry.getValue().toString());
		}
	}

	/**
	 * Uses regular File I/O to load and parse the configuration file identified by configFilname.
	 * The configuration file and referenced policies must be in the directory identified by
	 * policyDirectoryPath
	 * @param policyDirectoryPath the path of the directory that contains
	 * the referenced policies and the configuration file
	 * @param configFilename the name of the configuration file
	 * @throws IOException
	 */
	public ReferencePolicyFinderModule(String policyDirectoryPath, String configFilename) throws IOException {
		this.servletContext = null;
		this.policyDirectoryPath = policyDirectoryPath;
		
		String configPath = policyDirectoryPath + "/" + configFilename;
		log.info("configPath - " + configPath);
		InputStream configIs = new FileInputStream(configPath);
		configProps = new Properties();
		configProps.loadFromXML(configIs);
		for(Entry<Object, Object> entry : configProps.entrySet()) {
			log.info("key - " + entry.getKey().toString() + " - - value - " + entry.getValue().toString());
		}
	}

	/* (non-Javadoc)
	 * @see com.sun.xacml.finder.PolicyFinderModule#init(com.sun.xacml.finder.PolicyFinder)
	 */
	@Override
	public void init(PolicyFinder policyFinder) {
		this.policyFinder = policyFinder;
	}

	/* (non-Javadoc)
	 * @see com.sun.xacml.finder.PolicyFinderModule#findPolicy(java.net.URI, int, com.sun.xacml.VersionConstraints, com.sun.xacml.PolicyMetaData)
	 */
	@Override
	public PolicyFinderResult findPolicy(URI idReference, int type,
			VersionConstraints constraints, PolicyMetaData parentMetaData) {
		
		PolicyFinderResult result = null;

		InputStream policyIs = null;
		String policyPath = policyDirectoryPath + "/" + configProps.getProperty(idReference.toASCIIString());
		log.info("attempting to read file: " + policyPath + " to find policy: " + idReference.toASCIIString());
		
		AbstractPolicy policy = null;
		if(policyPath != null) {
			try {
				policyIs = getPolicyIs(policyPath);
				//policyIs = null;
				PolicyReader pRdr = new PolicyReader(policyFinder);
				policy = pRdr.readPolicy(policyIs);
			} catch (Exception e) {
				throw new RuntimeException(e);
			}
		}

		if(policy != null) {
			result = new PolicyFinderResult(policy);
		} else {
			result = new PolicyFinderResult();
		}

		return result;
	}

	/* (non-Javadoc)
	 * @see com.sun.xacml.finder.PolicyFinderModule#isIdReferenceSupported()
	 */
	@Override
	public boolean isIdReferenceSupported() {
		return true;
	}

	private InputStream getPolicyIs(String policyPath) throws Exception {
		if(servletContext != null)
			return servletContext.getResourceAsStream(policyPath);
		return new FileInputStream(policyPath);
	}
}
