/* Copyright 2011-2012, Georgia Institute of Technology. All rights reserved. */
package org.gtri.icl.iead.policy_guide.obligation;

import java.util.Map;

/**
 * This is the base class for all obligation handlers used in the Sample Application.
 * The fulfill method takes in an instance of an obligation via the argumentMap parameter-
 * a mapping from argument ID to an argument value. The validateArguments method is a
 * convenience method that can be used by implementations to ensure that the values
 * provided in the argumentMap parameter are of the appropriate Java Class.
 * @author Erin.Reddick AT gtri.gatech.edu
 *
 */
public abstract class ObligationHandler {

	/**
	 * Returns the Obligation ID handled by this class 
	 * @return the Obligation ID handled by this class
	 */
	abstract public String getHandledObligationId();

	/**
	 * Fulfills an obligation instance. An obligation instance is represented by
	 * the argumentMap parameter, which is a mapping from argument ID to an argument value.
	 * @param argumentMap mapping from argument ID to an argument value
	 * @param resource (optional) the requested data resource
	 * @return whether or not the obligation instance was fulfilled successfully
	 * @throws ObligationHandlerException
	 */
	abstract public boolean fulfill(Map<String, Object> argumentMap, Object resource) throws ObligationHandlerException;

	/**
	 * Validates an argument map of an obligation instance against argumentTypeMap.
	 * argumentTypeMap is a mapping from argument ID to the expected Java Class of argument values.
	 * argumentMap is a mapping from argument ID to an argument value.
	 * @param argumentTypeMap mapping from argument ID to the expected Java Class of argument values.
	 * @param argumentMap mapping from argument ID to an argument value
	 * @throws ObligationHandlerException
	 */
	public static void validateArguments(Map<String, Class<?>> argumentTypeMap, Map<String, Object> argumentMap)
			throws InvalidObligationArgumentException, ObligationHandlerException {
		int typeMapSize = argumentTypeMap.entrySet().size();
		int argMapSize = argumentMap.entrySet().size();
		if(typeMapSize != argMapSize)
			throw new ObligationHandlerException("Expected " + typeMapSize + " arguments. Received " + argMapSize + " arguments.");
		
		for(String argId : argumentTypeMap.keySet()) {
			Class<?> expectedClazz = argumentTypeMap.get(argId);
			Object argValue = argumentMap.get(argId);
			if(!(expectedClazz.isInstance(argValue)))
				throw new InvalidObligationArgumentException(argId, expectedClazz);
		}
	}
	
}
