/* Copyright 2011-2012, Georgia Institute of Technology. All rights reserved. */
package org.gtri.icl.iead.policy_guide.obligation;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Logger;

import oasis.names.tc.xacml._2_0.context.schema.os.RequestType;

import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import com.sun.xacml.BasicEvaluationCtx;
import com.sun.xacml.EvaluationCtx;
import com.sun.xacml.PolicyMetaData;
import com.sun.xacml.attr.AnyURIAttribute;
import com.sun.xacml.attr.AttributeDesignator;
import com.sun.xacml.attr.AttributeSelector;
import com.sun.xacml.attr.AttributeValue;
import com.sun.xacml.attr.BagAttribute;
import com.sun.xacml.attr.Base64BinaryAttribute;
import com.sun.xacml.attr.BooleanAttribute;
import com.sun.xacml.attr.DateAttribute;
import com.sun.xacml.attr.DateTimeAttribute;
import com.sun.xacml.attr.DoubleAttribute;
import com.sun.xacml.attr.IntegerAttribute;
import com.sun.xacml.attr.StringAttribute;
import com.sun.xacml.attr.TimeAttribute;
import com.sun.xacml.cond.Apply;
import com.sun.xacml.cond.Evaluatable;
import com.sun.xacml.cond.EvaluationResult;
import com.sun.xacml.finder.AttributeFinder;
import com.sun.xacml.finder.impl.SelectorModule;

/**
 * This class uses the SunXACML library to evaluate arguments
 * of XACML 2.0 obligations that are attribute references.
 * In XACML 3.0, this functionality is handled in the PDP.
 * @author Erin.Reddick AT gtri.gatech.edu
 *
 */
public class ObligationHandlerWrapper {

	/**
	 * The wrapped ObligationHandler
	 */
	private final ObligationHandler handler;
	
	/**
	 * The Logger used to write to the system log.
	 */
	private final Logger log = Logger.getLogger(ObligationHandlerWrapper.class.getName());
	
	/**
	 * Basic constructor that sets the wrapped ObligationHandler which must not be null.
	 * @param handler the wrapped ObligationHandler, must not be null;
	 */
	public ObligationHandlerWrapper(ObligationHandler handler) {
		if(handler == null)
			throw new IllegalArgumentException("handler must not be null");
		this.handler = handler;
	}

	/**
	 * Fulfills an obligation instance. An obligation instance is represented
	 * by the obligation parameter which is a DOM Element. xacmlRequest is
	 * the JAXB representation of the associated XACML request which is used
	 * to resolve attribute-reference arguments. This method resolves all
	 * attribute-references, and builds and passes an argument map to the
	 * wrapped ObligationHandler's fulfill method.
	 * @param obligation as a DOM Element
	 * @param xacmlRequest the associated XACML request
	 * @param resource (optional) the requested data resource
	 * @return the result of the invocation of the wrapped
	 * ObligationHandler's fulfill method.
	 * @throws ObligationHandlerException
	 */
	public boolean fulfill(Element obligation, RequestType xacmlRequest, Object resource) throws ObligationHandlerException {
		if(obligation == null)
			throw new IllegalArgumentException("obligation must not be null");
		if(xacmlRequest == null)
			throw new IllegalArgumentException("xacmlRequest must not be null");

		String obligationId = obligation.getAttribute("ObligationId");
		if(!(handler.getHandledObligationId().equals(obligationId)))
			throw new ObligationHandlerException("This handler cannot fulfill " + obligationId);

		Map<String, Object> arguments = new HashMap<String, Object>();

		try {
			AttributeFinder af = new AttributeFinder();
			List<Object> afModules = new ArrayList<Object>();
			afModules.add(new SelectorModule());
			af.setModules(afModules);
			EvaluationCtx evalCtx = new BasicEvaluationCtx(xacmlRequest, af);

			NodeList assignmentList = obligation.getElementsByTagNameNS("urn:oasis:names:tc:xacml:2.0:policy:schema:os", "AttributeAssignment");
			for(int i = 0; i < assignmentList.getLength(); i++) {
				Element assignment = (Element)assignmentList.item(i);
				String id = assignment.getAttribute("AttributeId");
				log.info("evaluating argument: " + id);

				Node child = assignment.getFirstChild();

				short nodeType = child.getNodeType();
				switch(nodeType) {
				case Node.TEXT_NODE:
					//System.out.println("String: " + (String)contentObj);
					String assignmentValue = child.getNodeValue();
					arguments.put(id, assignmentValue);
					break;

				case Node.ELEMENT_NODE:
					AttributeValue argValue = resolveExpressionArgument((Element)child, evalCtx);
					Object valueObj = null;
					Collection<AttributeValue> xacmlValues = null;
					Collection<Object> values = null;
					boolean isBag = false;

					if(argValue instanceof BagAttribute) {
						BagAttribute bagValue = (BagAttribute)argValue;
						isBag = true;
						xacmlValues = getBagCollection(bagValue);
						values = new ArrayList<Object>();
					} else {
						xacmlValues = new ArrayList<AttributeValue>();
						xacmlValues.add(argValue);
					}

					for(AttributeValue xacmlValue : xacmlValues) {

						if(xacmlValue.getType().equals(StringAttribute.identifierURI)) {
							valueObj = ((StringAttribute)xacmlValue).getValue();
						} else if(xacmlValue.getType().equals(BooleanAttribute.identifierURI)) {
							valueObj = ((BooleanAttribute)xacmlValue).getValue();
						} else if(xacmlValue.getType().equals(DateAttribute.identifierURI)) {
							valueObj = ((DateAttribute)xacmlValue).getValue();
						} else if(xacmlValue.getType().equals(DateTimeAttribute.identifierURI)) {
							valueObj = ((DateTimeAttribute)xacmlValue).getValue();
						} else if(xacmlValue.getType().equals(TimeAttribute.identifierURI)) {
							valueObj = ((TimeAttribute)xacmlValue).getValue();
						} else if(xacmlValue.getType().equals(AnyURIAttribute.identifierURI)) {
							valueObj = ((AnyURIAttribute)xacmlValue).getValue();
						} else if(xacmlValue.getType().equals(Base64BinaryAttribute.identifierURI)) {
							valueObj = ((Base64BinaryAttribute)xacmlValue).getValue();
						} else if(xacmlValue.getType().equals(IntegerAttribute.identifierURI)) {
							valueObj = ((IntegerAttribute)xacmlValue).getValue();
						} else if(xacmlValue.getType().equals(DoubleAttribute.identifierURI)) {
							valueObj = ((DoubleAttribute)xacmlValue).getValue();
						} else {
							valueObj = xacmlValue.encode();
						}
						
						if(isBag) {
							values.add(valueObj);
						}
					}

					if(isBag)
						arguments.put(id, values);
					else
						arguments.put(id, valueObj);
					break;

				default:
					throw new ObligationHandlerException("Did not find a valid argument value type");
				}

				System.out.println();
			}
		} catch(Exception e) {
			throw new ObligationHandlerException("Exception caught resolving obligation arguments", e);
		}

		return handler.fulfill(arguments, resource);
	}
	
	/**
	 * Extracts and returns the Collection object from the provided SunXACML BagAttribute
	 * @param bagValue a SunXACML BagAttribute
	 * @return the extracted Collection object
	 */
	@SuppressWarnings("unchecked")
	private Collection<AttributeValue> getBagCollection(BagAttribute bagValue) {
		return bagValue.getValue();
	}

	/**
	 * Resolves an attribute-reference obligation argument against the provided SunXACML
	 * EvaluationCtx. An EvaluationCtx is a SunXACML object that models XACML context
	 * elements provided in a XACML request as well as retrieved via a PIP. The
	 * attribute-reference is a DOM Element
	 * @param argument a DOM Element representation of a XACML obligation argument
	 * (&lt;AttributeAssignment&gt;)
	 * @param evalCtx the SunXACML EvaluationCtx
	 * @return the resolved argument value
	 * @throws Exception Exceptions thrown by the SunXACML library 
	 */
	private static AttributeValue resolveExpressionArgument(Element argument, EvaluationCtx evalCtx) throws Exception {
		Evaluatable evaluatable = null;
		PolicyMetaData pmd = new PolicyMetaData(2, 1);

		if(argument.getLocalName().equals("Apply")) {
			evaluatable = Apply.getInstance(argument, pmd, null);
		} else if(argument.getLocalName().equals("SubjectAttributeDesignator")) {
			evaluatable = AttributeDesignator.getInstance(argument, AttributeDesignator.SUBJECT_TARGET, pmd);
		} else if(argument.getLocalName().equals("ResourceAttributeDesignator")) {
			evaluatable = AttributeDesignator.getInstance(argument, AttributeDesignator.RESOURCE_TARGET, pmd);
		} else if(argument.getLocalName().equals("ActionAttributeDesignator")) {
			evaluatable = AttributeDesignator.getInstance(argument, AttributeDesignator.ACTION_TARGET, pmd);
		} else if(argument.getLocalName().equals("EnvironmentAttributeDesignator")) {
			evaluatable = AttributeDesignator.getInstance(argument, AttributeDesignator.ENVIRONMENT_TARGET, pmd);
		} else if(argument.getLocalName().equals("AttributeSelector")) {
			evaluatable = AttributeSelector.getInstance(argument, pmd);
		}

		EvaluationResult evalResult = evaluatable.evaluate(evalCtx);
		AttributeValue argValue = evalResult.getAttributeValue();
		return argValue;
	}
	
}
