/* Copyright 2011-2012, Georgia Institute of Technology. All rights reserved. */
package org.gtri.icl.iead.policy_guide.obligation.impl;

import java.text.DateFormat;
import java.util.Date;
import java.util.HashMap;
import java.util.Map;
import java.util.logging.Logger;

import org.gtri.icl.iead.policy_guide.obligation.ObligationHandler;
import org.gtri.icl.iead.policy_guide.obligation.ObligationHandlerException;

/**
 * Obligation Handler implementation for the LogValidAccess obligation
 * @author Erin.Reddick AT gtri.gatech.edu
 *
 */
public class LogValidAccessImpl extends ObligationHandler {

	/**
	 * The ID of the obligation that this class handles
	 */
	public static final String HANDLED_OBLIGATION_ID = "LogValidAccess";

	/**
	 * ID of the argument denoting the requested arrest record ID
	 */
	public static final String RECORDID_ID = "RecordId";
	
	/**
	 * ID of the argument denoting the ID of the requestor
	 */
	public static final String REQUESTORID_ID = "RequestorId";
	
	/**
	 * ID of the argument denoting the GFIPM Action Type
	 */
	public static final String ACTIONTYPE_ID = "ActionType";
	
	/**
	 * ID of the argument denoting the date-time of the access
	 */
	public static final String ACCESSDATETIME_ID = "AccessDateTime";
	
	/**
	 * ID of the argument denoting the date-time when the log entry "expires"
	 */
	public static final String EXPIRATIONDATETIME_ID = "ExpirationDateTime";

	/**
	 * This Logger instance used to write to the system log.
	 */
	private static final Logger log;

	/**
	 * Mapping from argument ID to the expected Java Class of the associated argument value 
	 */
	private static final Map<String, Class<?>> argumentTypeMap;

	static {
		log = Logger.getLogger(LogValidAccessImpl.class.getName());

		argumentTypeMap = new HashMap<String, Class<?>>();
		argumentTypeMap.put(RECORDID_ID, String.class);
		argumentTypeMap.put(REQUESTORID_ID, String.class);
		argumentTypeMap.put(ACTIONTYPE_ID, String.class);
		argumentTypeMap.put(ACCESSDATETIME_ID, Date.class);
		argumentTypeMap.put(EXPIRATIONDATETIME_ID, Date.class);
	}

	/* (non-Javadoc)
	 * @see org.gtri.icl.iead.policy_guide.obligation.ObligationHandler#fulfill(java.util.Map, java.lang.Object)
	 */
	@Override
	public boolean fulfill(Map<String, Object> argumentMap, Object resource)
			throws ObligationHandlerException {

		validateArguments(argumentTypeMap, argumentMap);
		
		String recordId = (String)argumentMap.get(RECORDID_ID);
		String requestorId = (String)argumentMap.get(REQUESTORID_ID);
		String actionType = (String)argumentMap.get(ACTIONTYPE_ID);
		Date accessDateTime = (Date)argumentMap.get(ACCESSDATETIME_ID);
		Date expDateTime = (Date)argumentMap.get(EXPIRATIONDATETIME_ID);

		try {
			String accessDtStr = DateFormat.getDateTimeInstance(DateFormat.FULL, DateFormat.FULL).format(accessDateTime);
			String expDtStr = DateFormat.getDateTimeInstance(DateFormat.FULL, DateFormat.FULL).format(expDateTime);
			
			String message = RECORDID_ID + ": " + recordId + "\n" +
					REQUESTORID_ID + ": " + requestorId + "\n" +
					ACTIONTYPE_ID + ": " + actionType + "\n" +
					ACCESSDATETIME_ID + ": " + accessDtStr + "\n" +
					EXPIRATIONDATETIME_ID + ": " + expDtStr;
			
			log.info(message);
			
		} catch(Exception e) {
			throw new ObligationHandlerException(e);
		}
		
		return true;
	}

	/* (non-Javadoc)
	 * @see org.gtri.icl.iead.policy_guide.obligation.ObligationHandler#getHandledObligationId()
	 */
	@Override
	public String getHandledObligationId() {
		return HANDLED_OBLIGATION_ID;
	}
}
