/* Copyright 2011-2012, Georgia Institute of Technology. All rights reserved. */
package org.gtri.icl.iead.policy_guide.obligation.impl;

import java.text.DateFormat;
import java.util.Date;
import java.util.HashMap;
import java.util.Map;
import java.util.logging.Logger;

import org.gtri.icl.iead.policy_guide.obligation.ObligationHandler;
import org.gtri.icl.iead.policy_guide.obligation.ObligationHandlerException;

/**
 * Obligation Handler implementation for the NotifyViaEmail obligation
 * @author Erin.Reddick AT gtri.gatech.edu
 *
 */
public class NotifyViaEmailImpl extends ObligationHandler {

	/**
	 * The ID of the obligation that this class handles
	 */
	public static final String HANDLED_OBLIGATION_ID = "NotifyViaEmail";
	
	/**
	 * ID of the argument denoting the email address to which to send the notification
	 */
	public static final String TARGETEMAIL_ID = "ArrestingOfficerEmailAddress";
	
	/**
	 * ID of the argument denoting the message of the notification
	 */
	public static final String MESSAGE_ID = "Message";
	
	/**
	 * ID of the argument denoting the accessed record ID
	 */
	public static final String RECORDID_ID = "RecordId";
	
	/**
	 * ID of the argument denoting the ID of the requestor
	 */
	public static final String REQUESTORID_ID = "RequestorId";
	
	/**
	 * ID of the argument denoting the name of the requestor's agency
	 */
	public static final String REQUESTORAGENCYNAME_ID = "RequestorAgencyName";
	
	/**
	 * ID of the argument denoting the date-time of the access
	 */
	public static final String ACCESSDATETIME_ID = "AccessDateTime";

	/**
	 * The Logger for writing to the system log
	 */
	private static final Logger log = Logger.getLogger(NotifyViaEmailImpl.class.getName());

	/**
	 * Mapping from argument ID to the expected Java Class of the associated argument value 
	 */
	private static final Map<String, Class<?>> argumentTypeMap;
	
	static {
		argumentTypeMap = new HashMap<String, Class<?>>();
		argumentTypeMap.put(TARGETEMAIL_ID, String.class);
		argumentTypeMap.put(MESSAGE_ID, String.class);
		argumentTypeMap.put(RECORDID_ID, String.class);
		argumentTypeMap.put(REQUESTORID_ID, String.class);
		argumentTypeMap.put(REQUESTORAGENCYNAME_ID, String.class);
		argumentTypeMap.put(ACCESSDATETIME_ID, Date.class);
	}

	/* (non-Javadoc)
	 * @see org.gtri.icl.iead.policy_guide.obligation.ObligationHandler#fulfill(java.util.Map, java.lang.Object)
	 */
	@Override
	public boolean fulfill(Map<String, Object> arguments, Object resource) throws ObligationHandlerException {
		validateArguments(argumentTypeMap, arguments);
		
		String targetEmailAddress = (String)arguments.get(TARGETEMAIL_ID);
		String message = (String)arguments.get(MESSAGE_ID);
		String recordId = (String)arguments.get(RECORDID_ID);
		String requestorId = (String)arguments.get(REQUESTORID_ID);
		String requestorAgencyName = (String)arguments.get(REQUESTORAGENCYNAME_ID);
		Date accessDatetime = (Date)arguments.get(ACCESSDATETIME_ID);
		
		try {
			//accessDatetime = DateTimeAttribute.getInstance((String)arguments.get(ACCESSDATETIME_ID)).getValue();

			//convert the accessedDatetime to a string
			String accessedDatetimeStr = DateFormat.getDateInstance(DateFormat.FULL).format(accessDatetime);
			accessedDatetimeStr += " " + DateFormat.getTimeInstance(DateFormat.FULL).format(accessDatetime);

			log.info("args: \ntargetEmail: " + targetEmailAddress +
					"\nmessage: " + message +
					"\nrecordid: " + recordId +
					"\nrequestorid: " + requestorId +
					"\nrequestoragency: " + requestorAgencyName +
					"\ndatetime: " + accessedDatetimeStr);

			String processedMessage = message;
			processedMessage = processedMessage.replaceAll("\\[" + RECORDID_ID + "\\]", recordId);
			processedMessage = processedMessage.replaceAll("\\[" + REQUESTORID_ID + "\\]", requestorId);
			processedMessage = processedMessage.replaceAll("\\[" + REQUESTORAGENCYNAME_ID + "\\]", requestorAgencyName);
			processedMessage = processedMessage.replaceAll("\\[" + ACCESSDATETIME_ID + "\\]", accessedDatetimeStr);

			log.info("Fulfilling the NotifyViaEmail Obligation:\n" +
					"To: " + targetEmailAddress + "\n" +
					"Subject: Notification of Data Access\n" +
					"Message:\n" +
					processedMessage);

		} catch(Exception e) {
			throw new ObligationHandlerException(e);
		}

		return true;
	}

	/* (non-Javadoc)
	 * @see org.gtri.icl.iead.policy_guide.obligation.ObligationHandler#getHandledObligationId()
	 */
	@Override
	public String getHandledObligationId() {
		return HANDLED_OBLIGATION_ID;
	}
}
