/* Copyright 2011-2012, Georgia Institute of Technology. All rights reserved. */
package org.gtri.icl.iead.policy_guide.service;

import java.io.BufferedReader;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.HashMap;
import java.util.Map;

import com.sun.xacml.attr.AnyURIAttribute;
import com.sun.xacml.attr.Base64BinaryAttribute;
import com.sun.xacml.attr.BooleanAttribute;
import com.sun.xacml.attr.DateAttribute;
import com.sun.xacml.attr.StringAttribute;

/**
 * This class contains a Map of attribute identifiers to XACMLAttribute instances
 * for GFIPM User attributes.
 * @author Erin.Reddick AT gtri.gatech.edu
 *
 */
public class GFIPMUserAttributes {

	/**
	 * Mapping from GFIPM User attribute identifier to XACMLAttribute
	 */
	private static final Map<String, XACMLAttribute> map = new HashMap<String, XACMLAttribute>();

	/**
	 * Flag that denotes if the attribute map has been built.
	 * This is used to ensure that the map is only built once.
	 */
	private static boolean mapIsBuilt = false;
	
	/**
	 * An InputStream of the GFIPM User attribute dictionary text file.
	 */
	private InputStream fileStream;
	
	/**
	 * Mapping from GFIPM datatype to XACML datatype
	 */
	public static final Map<String, String> xacmlDatatypeMap = new HashMap<String, String>();

	static {
		xacmlDatatypeMap.put("Text", StringAttribute.identifier);
		xacmlDatatypeMap.put("Boolean", BooleanAttribute.identifier);
		xacmlDatatypeMap.put("Date", DateAttribute.identifier);
		xacmlDatatypeMap.put("Base-64 Binary", Base64BinaryAttribute.identifier);
		xacmlDatatypeMap.put("URI", AnyURIAttribute.identifier);
	}

	/**
	 * Basic constructor that sets the InputStream
	 * of the GFIPM User attribute dictionary text file.
	 * @param fileStream InputStream of the GFIPM User
	 * attribute dictionary text file
	 */
	public GFIPMUserAttributes(InputStream fileStream) {
		this.fileStream = fileStream;
	}
	
	/**
	 * Returns the attribute mapping from GFIPM User
	 * attribute identifier to XACMLAttribute.
	 * The map is built if it hasn't yet been built.
	 * @return the attribute mapping from GFIPM User
	 * attribute identifier to XACMLAttribute
	 */
	public Map<String, XACMLAttribute> getMap() {
		if(!mapIsBuilt) {
			try {
				BufferedReader bRdr = new BufferedReader(new InputStreamReader(fileStream));
				
				//first line is a header line
				bRdr.readLine();

				String line = null;
				while((line = bRdr.readLine()) != null) {
					String[] parts = line.split("\t");
					String uri = parts[4];
					String gfipmDt = parts[2];
					String xacmlDt = xacmlDatatypeMap.get(gfipmDt);
					if(xacmlDt == null)
						xacmlDt = StringAttribute.identifier;
					//System.out.println("gfipm attr uri: " + uri);
					XACMLAttribute attr = new XACMLAttribute(uri, xacmlDt);
					map.put(uri, attr);
				}
				
				mapIsBuilt = true;
			} catch(Exception e) {
				System.out.println("in getMap() " + e.getMessage());
				throw new RuntimeException(e.getMessage(), e);
			}
		}

		return map;
	}
}
