/* Copyright 2011-2012, Georgia Institute of Technology. All rights reserved. */
package org.gtri.icl.iead.policy_guide.service;

import java.io.BufferedReader;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.StringReader;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.InvalidPropertiesFormatException;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;
import java.util.logging.Logger;

import javax.annotation.Resource;
import javax.jws.WebService;
import javax.security.auth.Subject;
import javax.servlet.ServletContext;
import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBElement;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Marshaller;
import javax.xml.bind.Unmarshaller;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamReader;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import javax.xml.ws.WebServiceContext;
import javax.xml.ws.handler.MessageContext;
import javax.xml.ws.soap.MTOM;

import oasis.names.tc.xacml._2_0.context.schema.os.ActionType;
import oasis.names.tc.xacml._2_0.context.schema.os.AttributeType;
import oasis.names.tc.xacml._2_0.context.schema.os.AttributeValueType;
import oasis.names.tc.xacml._2_0.context.schema.os.DecisionType;
import oasis.names.tc.xacml._2_0.context.schema.os.EnvironmentType;
import oasis.names.tc.xacml._2_0.context.schema.os.RequestType;
import oasis.names.tc.xacml._2_0.context.schema.os.ResourceContentType;
import oasis.names.tc.xacml._2_0.context.schema.os.ResourceType;
import oasis.names.tc.xacml._2_0.context.schema.os.ResponseType;
import oasis.names.tc.xacml._2_0.context.schema.os.ResultType;
import oasis.names.tc.xacml._2_0.context.schema.os.StatusCodeType;
import oasis.names.tc.xacml._2_0.context.schema.os.StatusType;
import oasis.names.tc.xacml._2_0.context.schema.os.SubjectType;

import org.gtri.icl.iead.policy_guide.arrest_exchange.ArrestReportType;
import org.gtri.icl.iead.policy_guide.arrest_extension.ArrestType;
import org.gtri.icl.iead.policy_guide.msg.ExceptionType;
import org.gtri.icl.iead.policy_guide.msg.ReadRequestType;
import org.gtri.icl.iead.policy_guide.msg.ReadResponseType;
import org.gtri.icl.iead.policy_guide.obligation.ObligationHandler;
import org.gtri.icl.iead.policy_guide.obligation.ObligationHandlerException;
import org.gtri.icl.iead.policy_guide.obligation.ObligationHandlerWrapper;
import org.gtri.icl.iead.policy_guide.sunxacml.ConfigurationStore;
import org.gtri.icl.iead.policy_guide.ws.FaultMsg;
import org.gtri.icl.iead.policy_guide.ws.SamplePortType;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;

import com.sun.xacml.PDP;
import com.sun.xacml.ParsingException;
import com.sun.xacml.UnknownIdentifierException;
import com.sun.xacml.attr.DateTimeAttribute;
import com.sun.xacml.attr.StringAttribute;
import com.sun.xacml.ctx.ResponseCtx;
import com.sun.xml.wss.SubjectAccessor;
import com.sun.xml.wss.XWSSecurityException;
import com.sun.xml.wss.saml.internal.saml20.jaxb20.AssertionType;
import com.sun.xml.wss.saml.internal.saml20.jaxb20.AttributeStatementType;
import com.sun.xml.wss.saml.internal.saml20.jaxb20.NameIDType;
import com.sun.xml.wss.saml.internal.saml20.jaxb20.StatementAbstractType;
import com.sun.xml.wss.saml.util.SAMLUtil;

/**
 * This contains the implementation of the web service
 * of the Sample Application.
 * @author Erin.Reddick AT gtri.gatech.edu
 *
 */
@MTOM
@WebService(serviceName = "SampleWebService",
portName = "SamplePort",
endpointInterface = "org.gtri.icl.iead.policy_guide.ws.SamplePortType",
targetNamespace = "http://iead.icl.gtri.org/policy-guide/ws",
wsdlLocation = "WEB-INF/wsdl/SampleWebserviceImpl.wsdl")
public class SampleWebServiceImpl implements SamplePortType {

	/**
	 * Default path to the SunXACML configuration file
	 */
	public static final String DEFAULT_SUNXACML_CONFIG_PATH = "/META-INF/policies/SunXACMLConfig.xml";

	/**
	 * Actual path to the SunXACML configuration file
	 */
	private final String sunxacmlConfigPath;
	
	/**
	 * Default path to the Obligation Handlers configuration file
	 */
	public static final String DEFAULT_OBLIGATION_HANDLER_CONFIG_PATH = "/META-INF/ObligationHandlerConfig.xml";

	/**
	 * Actual path to the Obligation Handlers configuration file
	 */
	private final String obligationHandlerConfigPath;
	
	/**
	 * Default path to the text file containing the GFIPM User attribute dictionary
	 */
	private static final String DEFAULT_GFIPM_USER_FILEPATH = "/META-INF/gfipm-metadata-2.0-user.txt";

	/**
	 * Actual path to the text file containing the GFIPM User attribute dictionary
	 */
	private final String gfipmUserFilePath;
	
	/**
	 * Default path to the directory containing the records.
	 */
	public static final String DEFAULT_DATA_DIRECTORY_PATH = "/META-INF/records";

	/**
	 * Actual path to the directory containing the records.
	 */
	private final String dataDirectoryPath;
	
	/**
	 * Value for the XACML "OK" Status
	 */
	public static final String XACML_STATUS_OK = "urn:oasis:names:tc:xacml:1.0:status:ok";

	/**
	 * The WebServiceContext which will be injected by the J2EE container
	 * or the test suite
	 */
	@Resource
    protected WebServiceContext wsContext;

	/**
	 * The SunXACML PDP
	 */
	private PDP pdp;

	/**
	 * Properties object for the Obligation Handlers configuration file
	 */
	private Properties obligationHandlerProperties;
	
	/**
	 * The ServletContext which will be extracted from the WebServiceContext
	 */
	private ServletContext servletContext;
	
	/**
	 * Mapping from Obligation ID to the associated ObligationHandlerWrapper.
	 * The ObligationHandlerWrapper wraps an ObligationHandler
	 */
	private Map<String, ObligationHandlerWrapper> obligationHandlerMap;

	/**
	 * Contains a Map from GFIPM User attribute identifier
	 * to XACMLAttribute instance
	 */
	private GFIPMUserAttributes gfipmUserAttributes;
	
	/**
	 * The Logger used for writing messages to the system log
	 */
	private Logger log = Logger.getLogger(SampleWebServiceImpl.class.getName());

	/**
	 * The DataAdapter used to read arrest records
	 */
	private DataAdapter dataAdapter;
	
	/**
	 * Uses the default paths.
	 */
	public SampleWebServiceImpl() {
		sunxacmlConfigPath = DEFAULT_SUNXACML_CONFIG_PATH;
		obligationHandlerConfigPath = DEFAULT_OBLIGATION_HANDLER_CONFIG_PATH;
		gfipmUserFilePath = DEFAULT_GFIPM_USER_FILEPATH;
		dataDirectoryPath = DEFAULT_DATA_DIRECTORY_PATH;
	}

	/**
	 * Takes in the paths to use
	 * @param sunxacmlConfigPath
	 * @param obligationHandlerConfigPath
	 * @param gfipmUserFilePath
	 * @param dataDirectoryPath
	 */
	public SampleWebServiceImpl(String sunxacmlConfigPath, String obligationHandlerConfigPath,
			String gfipmUserFilePath, String dataDirectoryPath) {
		this.sunxacmlConfigPath = sunxacmlConfigPath;
		this.obligationHandlerConfigPath = obligationHandlerConfigPath;
		this.gfipmUserFilePath = gfipmUserFilePath;
		this.dataDirectoryPath = dataDirectoryPath;
	}
	
	/**
	 * Extracts the servletContext from the wsContext
	 */
	public void initServletContext() {
		servletContext = (ServletContext) wsContext.getMessageContext().get(MessageContext.SERVLET_CONTEXT);
	}

	/**
	 * Parses the Obligation Handlers configuration file and builds the
	 * obligationHandlerMap.
	 * @throws InvalidPropertiesFormatException
	 * @throws IOException
	 */
	public void initObligationHandlers() throws InvalidPropertiesFormatException, IOException {
		obligationHandlerMap = new HashMap<String, ObligationHandlerWrapper>();

		InputStream configIs = servletContext.getResourceAsStream(obligationHandlerConfigPath);
		obligationHandlerProperties = new Properties();
		obligationHandlerProperties.loadFromXML(configIs);
		
		for(Object keyObj : obligationHandlerProperties.keySet()) {
			String obligationId = (String)keyObj;
			String obligationClazzName = (String)obligationHandlerProperties.get(keyObj);
			try {
				Class<?> obligationClazz = Class.forName(obligationClazzName);
				ObligationHandler handler = (ObligationHandler)obligationClazz.newInstance();
				obligationHandlerMap.put(obligationId, new ObligationHandlerWrapper(handler));
			} catch(Exception e) {
				log.info("Exception caught when instantiating ObligationHandler with ObligationId: " + obligationId);
			}
		}
	}

	/**
	 * Initializes gfipmUserAttributes which contains XACML attribute info
	 * about GFIPM User attributes
	 */
	private void initMetadata() {
		InputStream is = servletContext.getResourceAsStream(gfipmUserFilePath);
		if(is == null)
			throw new RuntimeException("gfipm user metadata file input stream is null");
		gfipmUserAttributes = new GFIPMUserAttributes(is);
	}
	
	/**
	 * Initializes the pdp using the SunXACML configuration file
	 * specified by SUNXACML_CONFIG_PATH
	 * @throws IOException
	 * @throws ParsingException
	 * @throws UnknownIdentifierException
	 */
	private void initPDP() throws IOException, ParsingException, UnknownIdentifierException {
		InputStream configIS = servletContext.getResourceAsStream(sunxacmlConfigPath);
		ConfigurationStore cs = new ConfigurationStore(configIS, servletContext);
		pdp = new PDP(cs.getDefaultPDPConfig());
	}

	/**
	 * Initializes the dataAdapter using the specified data directory path
	 */
	private void initDataAdapter() {
		dataAdapter = new DataAdapter(dataDirectoryPath);
	}
	
	/**
	 * Implements the read web service interface.
	 * @param readRequest the ReadRequestType containing the input parameters
	 * to the read service
	 */
    @Override
    public ReadResponseType read(ReadRequestType readRequest) throws FaultMsg {
    	ReadResponseType serviceResponse = null;
    	ArrestReportType resultDoc = new ArrestReportType();

    	try {
    		initServletContext();
    		initMetadata();
    		initPDP();
    		initObligationHandlers();
    		initDataAdapter();
    		
        	String recordId = readRequest.getRecordId().getValue();

    		//retrieve saml assertion
    		AssertionType samlAssertion = getAssertionType();

    		//retrieve data
        	ArrestType arrestRecord = dataAdapter.retrieveRecord(servletContext, recordId);
    		
    		//create XACML request
    		RequestType xacmlRequest = buildXACMLRequest(samlAssertion, recordId, arrestRecord);

    		//get access control decision
    		ResponseCtx responseCtx = pdp.evaluate(xacmlRequest);
    		
    		//SunXACML evaluates a request into an instance of ResponseCtx, which encodes the response to an XML string with no namespace info.
    		//We need to add in the namespace info.
    		String responseXmlStr = getResponseXmlStr(responseCtx);
    		responseXmlStr = addNsMetadataToResponseStr(responseXmlStr);
    		
    		//We need the response as a JaxbElement so we can attach it to the arrest report
    		ResponseType xacmlResponse = getJaxbResponse(responseXmlStr);
    		
    		//We need the Obligations as a DOM Element for obligation handling
    		Element obligationsElement = getObligationsAsElement(responseXmlStr);

    		//add the XACML request and response to the result document
    		//for demo purposes
    		oasis.names.tc.xacml._2_0.context.schema.os.ObjectFactory xacmlCtxOf = new oasis.names.tc.xacml._2_0.context.schema.os.ObjectFactory();
    		JAXBElement<RequestType> jaxbXacmlRequest = xacmlCtxOf.createRequest(xacmlRequest);
    		JAXBElement<ResponseType> jaxbXacmlResponse = xacmlCtxOf.createResponse(xacmlResponse);
    		jaxbXacmlRequest.setValue(xacmlRequest);
    		jaxbXacmlResponse.setValue(xacmlResponse);
    		resultDoc.setRequest(jaxbXacmlRequest);
    		resultDoc.setResponse(jaxbXacmlResponse);

    		//there should only be 1 result
    		List<ResultType> results = xacmlResponse.getResult();
    		ResultType result = results.get(0);
    		StatusType status = result.getStatus();
    		StatusCodeType statusCode = status.getStatusCode();
    		String code = statusCode.getValue();

    		//check result status for ok
    		if(!(code.equals(XACML_STATUS_OK)))
    			throw new FaultMsg("XACML Status: " + code, new ExceptionType());

    		//check decision for indeterminate
    		DecisionType decision = result.getDecision();
    		if(decision.equals(DecisionType.INDETERMINATE))
    			throw new FaultMsg("XACML Decision: " + decision, new ExceptionType());
    		
    		//process obligations
    		//processObligations(...) will throw a runtime exception if an obligation cannot be fulfilled
    		processObligations(xacmlRequest, obligationsElement, arrestRecord);
    		
    		//enforce decision
    		if(decision.equals(DecisionType.PERMIT)) {
    			//add the data to the result doc
    			resultDoc.setArrest(ObjectFactoryFactory.extOF.createArrest(arrestRecord));
    		} //otherwise, we deny access and do not add the data to the result doc
    		
    		//create the WS response
    		serviceResponse = new ReadResponseType();
    		serviceResponse.setArrestReport(ObjectFactoryFactory.exchOF.createArrestReport(resultDoc));
    		
    	} catch(Exception ex) {
    		log.fine("Exception caught in ReadResponseType: " + ex.getLocalizedMessage());
    		ex.printStackTrace(System.out);
    		throw new FaultMsg(ex.getLocalizedMessage(), new ExceptionType(), ex);
    	}

        return serviceResponse;
    }

    /**
     * Extracts the GFIPM SAML assertion that was provided to the web service
     * by the Web Service Consumer.
     * @return The GFIPM SAML assertion as a JAXB AssertionType 
     * @throws XWSSecurityException
     * @throws XMLStreamException
     * @throws TransformerException
     * @throws JAXBException
     */
	private AssertionType getAssertionType() throws XWSSecurityException, XMLStreamException, TransformerException, JAXBException {
		//get SAML attributes
		Subject authnSubj = SubjectAccessor.getRequesterSubject(wsContext);
		Set<Object> set = authnSubj.getPublicCredentials();
		Element samlAssertionElement = null;
		
		for (Object obj : set) {
			if (obj instanceof XMLStreamReader) {
				//System.out.println("obj is an instance of the XMLStreamReader");
				XMLStreamReader reader = (XMLStreamReader) obj;
				//To create a DOM Element representing the Assertion :
				samlAssertionElement = SAMLUtil.createSAMLAssertion(reader);
				break;
			} else if (obj instanceof Element) {
				//System.out.println("obj is an instance of the Element");
				samlAssertionElement = (Element) obj;
				break;
			} else if (obj instanceof JAXBElement<?>) {
				@SuppressWarnings("unchecked")
				JAXBElement<AssertionType> jaxbEl = (JAXBElement<AssertionType>)obj;
				return jaxbEl.getValue();
			} else {
				System.out.println("Well then, what do I have? " + obj.getClass().getCanonicalName());
			}
		}

		if(samlAssertionElement == null)
			System.out.println("samlAssertionElement is null");
		else
			dumpDomNode(samlAssertionElement);
		
		//unmarshal the samlAssertion
		JAXBContext jc = JAXBContext.newInstance(com.sun.xml.wss.saml.internal.saml20.jaxb20.ObjectFactory.class);
		Unmarshaller u = jc.createUnmarshaller();
		Object unmarshalledObj = u.unmarshal(samlAssertionElement);
		if(unmarshalledObj == null)
			throw new RuntimeException("SAML Assertion was null");
		//System.out.println("unmarshalledObj: " + unmarshalledObj);
		if(!(unmarshalledObj instanceof JAXBElement<?>))
			throw new RuntimeException("expected instanceof JAXBElement<?>");
		
		//System.out.println("unmarshalledObj is instance of JAXBElement<?>");
		@SuppressWarnings("unchecked")
		JAXBElement<AssertionType> jaxbSamlAssertion = (JAXBElement<AssertionType>)unmarshalledObj;
		AssertionType samlAssertion = jaxbSamlAssertion.getValue();
		//System.out.println(samlAssertion);
		
		return samlAssertion;
	}

	/**
	 * Uses the Obligation Handlers to handle a set of obligations. 
	 * @param request the XACML request as a JAXB RequestType 
	 * @param obligationsElement the set of obligations as a DOM Element
	 * @param arrestRecord the requested arrest data record as a JAXB ArrestType
	 * @throws ObligationHandlerException
	 */
    private void processObligations(RequestType request, Element obligationsElement, ArrestType arrestRecord) throws ObligationHandlerException {
    	if(obligationsElement == null) {
    		log.info("obligations element is null");
    		return;
    	}
    	
    	NodeList obligationList = obligationsElement.getElementsByTagNameNS("urn:oasis:names:tc:xacml:2.0:policy:schema:os", "Obligation");
    	for(int i = 0; i < obligationList.getLength(); i++) {
    		Element obligation = (Element)obligationList.item(i);
    		String obligationId = obligation.getAttribute("ObligationId");
    		ObligationHandlerWrapper wrapper = obligationHandlerMap.get(obligationId);
    		if(wrapper == null)
    			throw new RuntimeException("No Handler found for Obligation ID: " + obligationId);

    		boolean fulfilled = false;
    		fulfilled = wrapper.fulfill(obligation, request, arrestRecord);

    		if(!fulfilled)
    			throw new ObligationHandlerException("Obligation with ID: " + obligationId + " not fulfilled.");
    	}
    }

    /**
     * Convenience method for writing the contents of a DOM Node
     * @param node the DOM node to dump
     * @throws TransformerException
     */
    private static void dumpDomNode(Node node) throws TransformerException {
        System.out.println("==== DebugUtil.dumpDomNode(...) Start ====");
        DOMSource domSource = new DOMSource(node);
        TransformerFactory tf = TransformerFactory.newInstance();
        Transformer xform = null;
        xform = tf.newTransformer();
        xform.transform(domSource, new StreamResult(System.out));
        System.out.println();
        System.out.println("==== DebugUtil.dumpDomNode(...) End ====");
    }

    /**
     * Builds a XACML request from the provided SAML assertion,
     * record id and arrest data record
     * @param samlAssertion the GFIPM SAML assertion as a JAXB AssertType
     * @param recordId the identifier of the requested arrest record-
     * set as the XACML resource ID
     * @param arrestRecord the requested arrest record as a JAXB ArrestType
     * @return a JAXB RequestType
     * @throws JAXBException
     * @throws ParserConfigurationException
     */
	private RequestType buildXACMLRequest(AssertionType samlAssertion, String recordId, ArrestType arrestRecord) throws JAXBException, ParserConfigurationException {
		//capture the current date-time
    	Date currentDatetime = new Date();
    	
    	//marshal arrestRecord to a DOM Node
    	Element arrestRecordElement = null;
    	if(arrestRecord != null) {
    		JAXBContext jc = JAXBContext.newInstance(ObjectFactoryFactory.extOF.getClass(),
    				ObjectFactoryFactory.aiOF.getClass(),
    				ObjectFactoryFactory.niemOF.getClass(),
    				ObjectFactoryFactory.niemProxyOF.getClass(),
    				ObjectFactoryFactory.niemStructOF.getClass(),
    				ObjectFactoryFactory.ctxOF.getClass(),
    				ObjectFactoryFactory.policyOF.getClass());

    		Marshaller m = jc.createMarshaller();
    		DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
    		dbf.setIgnoringComments(true);
    		dbf.setNamespaceAware(true);
    		DocumentBuilder db = dbf.newDocumentBuilder();
    		Document doc = db.newDocument();
    		m.marshal(ObjectFactoryFactory.extOF.createArrest(arrestRecord), doc);
    		arrestRecordElement = doc.getDocumentElement();
    	}
    	
		RequestType xacmlRequest = new RequestType();
		
		//***********BUILD SUBJECT******************/
		SubjectType subject = new SubjectType();

		//extract issuer
		NameIDType issuerType = samlAssertion.getIssuer();
		String issuer = issuerType.getValue();
		
		//extract saml-attributes and create xacml-subject-attributes
		List<StatementAbstractType> stmtAbstTypes = samlAssertion.getStatementOrAuthnStatementOrAuthzDecisionStatement();
		for(StatementAbstractType stmtAbstType : stmtAbstTypes) {
			if(!(stmtAbstType instanceof AttributeStatementType))
				continue;
			
			AttributeStatementType attrStmtType = (AttributeStatementType)stmtAbstType;
			List<Object> attrObjs = attrStmtType.getAttributeOrEncryptedAttribute();

			for(Object attrObj : attrObjs) {
				if(!(attrObj instanceof com.sun.xml.wss.saml.internal.saml20.jaxb20.AttributeType))
					continue;
				
				com.sun.xml.wss.saml.internal.saml20.jaxb20.AttributeType samlAttr = (com.sun.xml.wss.saml.internal.saml20.jaxb20.AttributeType)attrObj;
				//System.out.println(attrType.getName());
				String xacmlAttrId = samlAttr.getName();
				
				AttributeType attribute = new AttributeType();
				attribute.setIssuer(issuer);
				attribute.setAttributeId(xacmlAttrId);
				
				String datatype = null;
				XACMLAttribute gfipmAttrInfo = gfipmUserAttributes.getMap().get(xacmlAttrId);
				if(gfipmAttrInfo != null) {
					//this is a standard GFIPM User Attribute
					//get the appropriate datatype
					datatype = gfipmAttrInfo.getDatatype();
				}
				if(datatype == null)
					datatype = StringAttribute.identifier;
				attribute.setDataType(datatype);
				
				//attribute values
				for(Object samlAttrValueObj : samlAttr.getAttributeValue()) {
					String valueContent = null;
					
					if(samlAttrValueObj instanceof Element) {
						//since in SAML, <AttributeValue> is of XML type "anyType",
						//then JAX-B will map it to java.lang.Object.
						//but the Java Object will be an instanceof org.w3c.dom.Element
						//thus we need to manually parse the DOM Element
						Element samlAttrValueElement = (Element)samlAttrValueObj;
						valueContent = samlAttrValueElement.getTextContent();
					} else if(samlAttrValueObj instanceof String) {
						valueContent = (String)samlAttrValueObj;
					} else {
						throw new RuntimeException("Failed to parse SAML attribute value for: " + xacmlAttrId);
					}
					
					/*
					if(datatype.equals("http://www.w3.org/2001/XMLSchema#boolean")) {
						Boolean booleanValue = new Boolean(valueContent);
						if(booleanValue) {
							valueContent = "True";
						} else {
							valueContent = "False";
						}
					}
					*/
					
					AttributeValueType attributeValue = new AttributeValueType();
					attributeValue.getContent().add(valueContent);
					attribute.getAttributeValue().add(attributeValue);
				}
				
				subject.getAttribute().add(attribute);
			}
		}
		//***********DONE BUILDING SUBJECT******************/

		xacmlRequest.getSubject().add(subject);
		
		//resource
		ResourceType resource = new ResourceType();

		//**************BUILD RESOURCE***************/
		//resource-id
		AttributeType resourceIdAttribute = new AttributeType();
		resourceIdAttribute.setAttributeId(XACMLStandardAttribute.RESOURCE_ID.getAttribute().getUri());
		resourceIdAttribute.setDataType(XACMLStandardAttribute.RESOURCE_ID.getAttribute().getDatatype());
		AttributeValueType resourceIdAttributeValue = new AttributeValueType();
		resourceIdAttributeValue.getContent().add(recordId);
		resourceIdAttribute.getAttributeValue().add(resourceIdAttributeValue);
		resource.getAttribute().add(resourceIdAttribute);

		//criminal history data indicator
		AttributeType chdiAttr = new AttributeType();
		chdiAttr.setAttributeId(GFIPMAttribute.CRIMINAL_HISTORY_DATA_INDICATOR.getAttribute().getUri());
		chdiAttr.setDataType(GFIPMAttribute.CRIMINAL_HISTORY_DATA_INDICATOR.getAttribute().getDatatype());
		AttributeValueType chdiAttrValue = new AttributeValueType();
		chdiAttrValue.getContent().add("true");
		chdiAttr.getAttributeValue().add(chdiAttrValue);
		resource.getAttribute().add(chdiAttr);
		
		//resource content
		ResourceContentType resourceContent = new ResourceContentType();
		//resourceContent.getContent().add(arOF.createArrestRecord(arrestRecord));
		if(arrestRecordElement != null)
			resourceContent.getContent().add(arrestRecordElement);
		resource.setResourceContent(resourceContent);
		//***************DONE BUILDING RESOURCE*************/
		
		xacmlRequest.getResource().add(resource);
		
		//action
		ActionType action = new ActionType();
		AttributeType actionIdAttribute = new AttributeType();
		actionIdAttribute.setAttributeId(GFIPMAttribute.ACTION_TYPE.getAttribute().getUri());
		actionIdAttribute.setDataType(GFIPMAttribute.ACTION_TYPE.getAttribute().getDatatype());
		AttributeValueType actionIdValue = new AttributeValueType();
		actionIdValue.getContent().add(GFIPMActionTypeCode.READ.getCode());
		actionIdAttribute.getAttributeValue().add(actionIdValue);
		action.getAttribute().add(actionIdAttribute);
		xacmlRequest.setAction(action);
		
		//environment
		EnvironmentType env = new EnvironmentType();
		AttributeType currentDatetimeAttribute = new AttributeType();
		currentDatetimeAttribute.setAttributeId(XACMLStandardAttribute.CURRENT_DATETIME.getAttribute().getUri());
		currentDatetimeAttribute.setDataType(XACMLStandardAttribute.CURRENT_DATETIME.getAttribute().getDatatype());
		AttributeValueType currentDatetimeValue = new AttributeValueType();
		currentDatetimeValue.getContent().add(new DateTimeAttribute(currentDatetime).encode());
		currentDatetimeAttribute.getAttributeValue().add(currentDatetimeValue);
		env.getAttribute().add(currentDatetimeAttribute);
		xacmlRequest.setEnvironment(env);

		return xacmlRequest;
	}

	/**
	 * Converts a ResponseCtx into its XML String representation.
	 * This string representation won't have any XML namespace info
	 * since SunXACML doesn't include XML namespace info in its responses
	 * @param respCtx the ResponseCtx to convert
	 * @return XML String representation of respCtx
	 * @throws JAXBException
	 */
	private String getResponseXmlStr(ResponseCtx respCtx) throws JAXBException {
		ByteArrayOutputStream respBaos = new ByteArrayOutputStream();
		respCtx.encode(respBaos);
		String respCtxStr = respBaos.toString();
		
		//System.out.println("response str: " + respCtxStr);
		
		return respCtxStr;
	}

	/**
	 * Adds XML namespace info to the provided XACML response string
	 * @param respCtxStr the XACML response string
	 * @return the XACML response string with XML namespace info added
	 */
	private String addNsMetadataToResponseStr(String respCtxStr) {
		//add namespace info, replace non-qualified elements with qualified elements
		
		Set<String> ctxElementNames = new HashSet<String>();
		ctxElementNames.add("Response");
		ctxElementNames.add("Result");
		ctxElementNames.add("Decision");
		ctxElementNames.add("Status");
		ctxElementNames.add("StatusCode");
		ctxElementNames.add("StatusMessage");
		ctxElementNames.add("StatusDetail");
		ctxElementNames.add("MissingAttributeDetail");
		//ctxElementNames.add("AttributeValue");
		
		Set<String> polElementNames = new HashSet<String>();
		polElementNames.add("Obligations");
		polElementNames.add("Obligation");
		polElementNames.add("AttributeAssignment");
		polElementNames.add("Apply");
		polElementNames.add("Function");
		polElementNames.add("AttributeValue");
		polElementNames.add("AttributeSelector");
		polElementNames.add("SubjectAttributeDesignator");
		polElementNames.add("ResourceAttributeDesignator");
		polElementNames.add("ActionAttributeDesignator");
		polElementNames.add("EnvironmentAttributeDesignator");
		
		String ctxNs = "urn:oasis:names:tc:xacml:2.0:context:schema:os";
		String polNs = "urn:oasis:names:tc:xacml:2.0:policy:schema:os";
		String ctxPrefix = "xacml-ctx";
		String polPrefix = "xacml";

		respCtxStr = respCtxStr.replace("<Response", "<" + ctxPrefix + ":" + "Response xmlns:"
				+ ctxPrefix + "=\"" + ctxNs + "\""
				+ " xmlns:" + polPrefix + "=\"" + polNs + "\"");

		for(String ctxElementName : ctxElementNames) {
			respCtxStr = respCtxStr.replace("<" + ctxElementName, "<" + ctxPrefix + ":" + ctxElementName);
			respCtxStr = respCtxStr.replace("</" + ctxElementName, "</" + ctxPrefix + ":" + ctxElementName);
		}


		for(String polElementName : polElementNames) {
			respCtxStr = respCtxStr.replace("<" + polElementName, "<" + polPrefix + ":" + polElementName);
			respCtxStr = respCtxStr.replace("</" + polElementName, "</" + polPrefix + ":" + polElementName);
		}
		
		//System.out.println("converted response str: " + respCtxStr);
		
		return respCtxStr;
	}
	
	/**
	 * Marshals a XACML response XML string (with appropriate XML namespace info)
	 * into a JAXB ResponseType object
	 * @param responseXmlStr the XACML response XML string
	 * @return the marshaled JAXB ResponseType object
	 * @throws JAXBException
	 */
	private ResponseType getJaxbResponse(String responseXmlStr) throws JAXBException {
		
		ByteArrayInputStream respBais = new ByteArrayInputStream(responseXmlStr.getBytes());
		
		JAXBContext jc = JAXBContext.newInstance(ResponseType.class.getPackage().getName());
		Unmarshaller u = jc.createUnmarshaller();
		Object unmarshalledObj = u.unmarshal(respBais);

		@SuppressWarnings("unchecked")
		JAXBElement<ResponseType> responseJaxbElement = (JAXBElement<ResponseType>)unmarshalledObj;

		ResponseType xacmlResponse = responseJaxbElement.getValue();
		
		return xacmlResponse;
	}

	/**
	 * Parses a XACML response XML string into a DOM Element
	 * @param responseXmlStr the XACML response XML string
	 * @return the XACML response as a DOM Element
	 * @throws Exception
	 */
	private Element getObligationsAsElement(String responseXmlStr) throws Exception {
		DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
		factory.setNamespaceAware(true);
		DocumentBuilder builder = factory.newDocumentBuilder();
		InputSource is = new InputSource(new StringReader(responseXmlStr));
		Document d = builder.parse(is);
		Element responseElement = d.getDocumentElement();

		/*
		TransformerFactory transFactory = TransformerFactory.newInstance();
		Transformer transformer = transFactory.newTransformer();
		StringWriter buffer = new StringWriter();
		transformer.setOutputProperty(OutputKeys.OMIT_XML_DECLARATION, "yes");
		transformer.transform(new DOMSource(responseElement),
		      new StreamResult(buffer));
		String str = buffer.toString();
		System.out.println("newstr: " + str);
		*/

		NodeList list = responseElement.getElementsByTagNameNS("urn:oasis:names:tc:xacml:2.0:context:schema:os", "Result");
		Element resultElement = (Element)list.item(0);
		return (Element)resultElement.getElementsByTagNameNS("urn:oasis:names:tc:xacml:2.0:policy:schema:os", "Obligations").item(0);
	}
	
	public static void main(String[] args) throws Exception {
		SampleWebServiceImpl impl = new SampleWebServiceImpl();
		
		String metaInfPath = "src" + File.separator + "main" + File.separator + "webapp" + File.separator + "META-INF";
		String gfipmUserAttrFilename = "gfipm-metadata-2.0-user.txt";
		String gfipmUserAttrPath = metaInfPath + File.separator + gfipmUserAttrFilename;
		InputStream gfipmUserAttrIS = new FileInputStream(gfipmUserAttrPath);
		impl.gfipmUserAttributes = new GFIPMUserAttributes(gfipmUserAttrIS);

		String samlAssertionFilename = "saml_assertion_user05.xml";
		String samlAssertionPath = "src" + File.separator + "test" + File.separator + samlAssertionFilename;
		File samlAssertionFile = new File(samlAssertionPath);
		JAXBContext samlJc = JAXBContext.newInstance(com.sun.xml.wss.saml.internal.saml20.jaxb20.AssertionType.class);
		Unmarshaller samlU = samlJc.createUnmarshaller();
		@SuppressWarnings("unchecked")
		JAXBElement<AssertionType> assertionJaxbEl = (JAXBElement<AssertionType>)samlU.unmarshal(samlAssertionFile);
		AssertionType assertion = assertionJaxbEl.getValue();

		String recordId = "Record-2";
		String recordPath = metaInfPath + File.separator + "records" +
				File.separator + recordId + ".xml";
		File recordFile = new File(recordPath);
		JAXBContext arrJc = JAXBContext.newInstance(ObjectFactoryFactory.extOF.getClass(),
				ObjectFactoryFactory.ctxOF.getClass());
		Unmarshaller arrU = arrJc.createUnmarshaller();
		@SuppressWarnings("unchecked")
		JAXBElement<ArrestType> arrestJaxbEl = (JAXBElement<ArrestType>)arrU.unmarshal(recordFile);
		ArrestType arrest = arrestJaxbEl.getValue();
		
		RequestType request = impl.buildXACMLRequest(assertion, recordId, arrest);
		
		//String responseXmlStr = "<xacml-ctx:Response xmlns:xacml-ctx=\"urn:oasis:names:tc:xacml:2.0:context:schema:os\" xmlns:xacml=\"urn:oasis:names:tc:xacml:2.0:policy:schema:os\"> <xacml-ctx:Result ResourceId=\"Record-2\"> <xacml-ctx:Decision>Permit</xacml-ctx:Decision> <xacml-ctx:Status> <xacml-ctx:StatusCode Value=\"urn:oasis:names:tc:xacml:1.0:status:ok\"/> </xacml-ctx:Status> <xacml:Obligations> <xacml:Obligation FulfillOn=\"Permit\" ObligationId=\"NotifyViaEmail\"> <xacml:AttributeAssignment AttributeId=\"ArrestingOfficerEmailAddress\" DataType=\"http://www.w3.org/2001/XMLSchema#string\"><xacml:Apply FunctionId=\"urn:oasis:names:tc:xacml:1.0:function:string-one-and-only\"> <xacml:AttributeSelector xmlns:ctx=\"urn:oasis:names:tc:xacml:2.0:context:schema:os\" xmlns:ext=\"http://iead.icl.gtri.org/policy-guide/arrest-extension\" xmlns:fips_5-2=\"http://niem.gov/niem/fips_5-2/2.0\" xmlns:j=\"http://niem.gov/niem/domains/jxdm/4.1\" xmlns:nc=\"http://niem.gov/niem/niem-core/2.0\" xmlns:niem-xsd=\"http://niem.gov/niem/proxy/xsd/2.0\" xmlns:s=\"http://niem.gov/niem/structures/2.0\" DataType=\"http://www.w3.org/2001/XMLSchema#string\" RequestContextPath=\"//ctx:Resource/ctx:ResourceContent/ext:Arrest/ext:ArrestOfficial/nc:ContactEmailID/text()\"/> </xacml:Apply></xacml:AttributeAssignment> <xacml:AttributeAssignment AttributeId=\"Message\" DataType=\"http://www.w3.org/2001/XMLSchema#string\">[ResourceId] has been accessed by [RequestorId] of [RequestorAgencyName] at [AccessDateTime].</xacml:AttributeAssignment> <xacml:AttributeAssignment AttributeId=\"RecordId\" DataType=\"http://www.w3.org/2001/XMLSchema#string\"><xacml:Apply FunctionId=\"urn:oasis:names:tc:xacml:1.0:function:string-one-and-only\"> <xacml:ResourceAttributeDesignator AttributeId=\"urn:oasis:names:tc:xacml:1.0:resource:resource-id\" DataType=\"http://www.w3.org/2001/XMLSchema#string\"/> </xacml:Apply></xacml:AttributeAssignment> <xacml:AttributeAssignment AttributeId=\"RequestorId\" DataType=\"http://www.w3.org/2001/XMLSchema#string\"><xacml:Apply FunctionId=\"urn:oasis:names:tc:xacml:1.0:function:string-one-and-only\"> <xacml:SubjectAttributeDesignator AttributeId=\"gfipm:2.0:user:FederationId\" DataType=\"http://www.w3.org/2001/XMLSchema#string\"/> </xacml:Apply></xacml:AttributeAssignment> <xacml:AttributeAssignment AttributeId=\"RequestorAgencyName\" DataType=\"http://www.w3.org/2001/XMLSchema#string\"><xacml:Apply FunctionId=\"urn:oasis:names:tc:xacml:1.0:function:string-one-and-only\"> <xacml:SubjectAttributeDesignator AttributeId=\"gfipm:2.0:user:EmployerName\" DataType=\"http://www.w3.org/2001/XMLSchema#string\"/> </xacml:Apply></xacml:AttributeAssignment> <xacml:AttributeAssignment AttributeId=\"AccessDateTime\" DataType=\"http://www.w3.org/2001/XMLSchema#string\"><xacml:Apply FunctionId=\"urn:oasis:names:tc:xacml:1.0:function:dateTime-one-and-only\"> <xacml:EnvironmentAttributeDesignator AttributeId=\"urn:oasis:names:tc:xacml:1.0:environment:current-dateTime\" DataType=\"http://www.w3.org/2001/XMLSchema#dateTime\"/> </xacml:Apply></xacml:AttributeAssignment> </xacml:Obligation> <xacml:Obligation FulfillOn=\"Permit\" ObligationId=\"LogValidAccess\"> <xacml:AttributeAssignment AttributeId=\"RequestorId\" DataType=\"http://www.w3.org/2001/XMLSchema#string\"><xacml:Apply FunctionId=\"urn:oasis:names:tc:xacml:1.0:function:string-one-and-only\"> <xacml:SubjectAttributeDesignator AttributeId=\"gfipm:2.0:user:FederationId\" DataType=\"http://www.w3.org/2001/XMLSchema#string\"/> </xacml:Apply></xacml:AttributeAssignment> <xacml:AttributeAssignment AttributeId=\"ResourceId\" DataType=\"http://www.w3.org/2001/XMLSchema#string\"><xacml:Apply FunctionId=\"urn:oasis:names:tc:xacml:1.0:function:string-one-and-only\"> <xacml:ResourceAttributeDesignator AttributeId=\"urn:oasis:names:tc:xacml:1.0:resource:resource-id\" DataType=\"http://www.w3.org/2001/XMLSchema#string\"/> </xacml:Apply></xacml:AttributeAssignment> <xacml:AttributeAssignment AttributeId=\"ActionType\" DataType=\"http://www.w3.org/2001/XMLSchema#string\"><xacml:Apply FunctionId=\"urn:oasis:names:tc:xacml:1.0:function:string-one-and-only\"> <xacml:ActionAttributeDesignator AttributeId=\"gfipm:2.0:action:ActionType\" DataType=\"http://www.w3.org/2001/XMLSchema#string\"/> </xacml:Apply></xacml:AttributeAssignment> <xacml:AttributeAssignment AttributeId=\"AccessDateTime\" DataType=\"http://www.w3.org/2001/XMLSchema#string\"><xacml:Apply FunctionId=\"urn:oasis:names:tc:xacml:1.0:function:dateTime-one-and-only\"> <xacml:EnvironmentAttributeDesignator AttributeId=\"urn:oasis:names:tc:xacml:1.0:environment:current-dateTime\" DataType=\"http://www.w3.org/2001/XMLSchema#dateTime\"/> </xacml:Apply></xacml:AttributeAssignment> <xacml:AttributeAssignment AttributeId=\"ExpirationDateTime\" DataType=\"http://www.w3.org/2001/XMLSchema#string\"><xacml:Apply FunctionId=\"urn:oasis:names:tc:xacml:1.0:function:dateTime-add-yearMonthDuration\"> <xacml:Apply FunctionId=\"urn:oasis:names:tc:xacml:1.0:function:dateTime-one-and-only\"> <xacml:EnvironmentAttributeDesignator AttributeId=\"urn:oasis:names:tc:xacml:1.0:environment:current-dateTime\" DataType=\"http://www.w3.org/2001/XMLSchema#dateTime\"/> </xacml:Apply> <xacml:AttributeValue DataType=\"http://www.w3.org/TR/2002/WD-xquery-operators-20020816#yearMonthDuration\">P60M</xacml:AttributeValue> </xacml:Apply> </xacml:AttributeAssignment> </xacml:Obligation> </xacml:Obligations> </xacml-ctx:Result> </xacml-ctx:Response>";
		String responseFilename = "Request-6_Sample-Application-Policy_Response.xml";
		String responsePath = "src" + File.separator + "test" + File.separator + responseFilename;
		File responseFile = new File(responsePath);
		BufferedReader responseRdr = new BufferedReader(new FileReader(responseFile));
		String responseXmlStr = "", line = null;
		
		while((line = responseRdr.readLine()) != null) {
			responseXmlStr += line;
		}
		
		responseXmlStr = impl.addNsMetadataToResponseStr(responseXmlStr);
		
		Element obligationsElement = impl.getObligationsAsElement(responseXmlStr);
		
		impl.processObligations(request, obligationsElement, arrest);
		
	}
}
